@@ -5,6 +5,7 @@ from tools_hjh.DBConn import QueryResults
 from tools_hjh.MemoryDB import MemoryDB
 from tools_hjh.Tools import line_merge_align, locatdate, merge_spaces, \
     analysis_hosts, locattime, remove_leading_space
+from numpy.core.defchararray import upper
 
 
 def main():
@@ -20,20 +21,20 @@ def main():
 
 class OracleTools:
     """ 用于Oracle的工具类 """
-
+    
     @staticmethod
     def desc(ora_conn, username, table):
-        """ 输出一个表的，列名、字段类型、主键、索引（索引所在的列而非索引名）和默认值 """
         username = username.upper()
         table = table.upper()
+        mess = 'table: ' + username.lower() + '.' + table.lower() + '\n'
         
-        # 列
+        # 列 类型 非空约束 默认值
         sql = '''
             select column_name, 
                 case 
-                    when data_type = 'VARCHAR2' or data_type = 'CHAR' or data_type = 'VARCHAR' then 
+                    when data_type = 'VARCHAR2' or data_type = 'CHAR' then 
                         data_type || '(' || data_length || ')'
-                    when data_type = 'NVARCHAR' or data_type = 'NVARCHAR2' then 
+                    when data_type = 'NVARCHAR2' then 
                         data_type || '(' || char_length || ')'
                     when data_type = 'NUMBER' and data_precision > 0 and data_scale > 0 then 
                         data_type || '(' || data_precision || ', ' || data_scale || ')'
@@ -42,47 +43,29 @@ class OracleTools:
                     when data_type = 'NUMBER' and data_precision = 0 and data_scale = 0 then 
                         data_type
                     else data_type 
-                end column_type
+                end column_type, nullable, data_default
             from dba_tab_cols where owner = ? and table_name = ? and column_name not like '%$%' order by column_id
         '''
-        tab = ''
         cols_ = ora_conn.run(sql, (username, table)).get_rows()
         lenNum = 0
         for col_ in cols_:
             if lenNum < len(col_[0]):
                 lenNum = len(col_[0])
         for col_ in cols_:
-            tablename = col_[0]
+            colstr = col_[0]
             typestr = col_[1]
-            spacesnum = lenNum - len(tablename) + 1
-            colstr = tablename + ' ' * (spacesnum) + typestr
-            tab = tab + colstr + '\n'
+            colstr = colstr + ' ' + typestr
+            if col_[2] == 'N':
+                nullable = ' not null'
+            else:
+                nullable = ''
+            if col_[3] != 'None':
+                data_default = ' default ' + col_[3]
+            else:
+                data_default = ''
+            mess = mess + 'column: ' + colstr.lower() + nullable + data_default + '\n'
             
-        # 主键
-        sql = '''
-            select constraint_name
-            , max(cols)
-            from (
-                select t.constraint_name
-                , to_char(
-                    wm_concat(t2.column_name) 
-                    over(partition by t.constraint_name order by t2.position)
-                ) cols
-                from dba_constraints t, dba_cons_columns t2
-                where t.owner = ?
-                and t.table_name = ?
-                and t.owner = t2.owner
-                and t.constraint_name = t2.constraint_name
-                and t.table_name = t2.table_name
-                and t.constraint_type = 'P'
-            ) group by constraint_name
-        '''
-        rss = ora_conn.run(sql, (username, table)).get_rows()
-        for rs in rss:
-            mess = 'pk: (' + rs[1] + ')'
-            tab = tab + mess + '\n'
-        
-        # 索引，包括普通索引和普通函数索引
+        # 索引 类型 列和列排序
         sql = '''
             select t.index_name
             , t.index_type
@@ -90,6 +73,7 @@ class OracleTools:
             , t3.column_expression
             , t2.descend
             , t2.column_position
+            , t.uniqueness
             from dba_indexes t, dba_ind_columns t2, dba_ind_expressions t3
             where t.table_owner = ? 
             and t.table_name = ? 
@@ -102,8 +86,6 @@ class OracleTools:
             and t2.index_name = t3.index_name(+)
             and t2.table_name = t3.table_name(+)
             and t2.column_position = t3.column_position(+)
-            and t.uniqueness = 'NONUNIQUE'
-            and t.index_type in('FUNCTION-BASED NORMAL', 'NORMAL')
             order by t2.column_position
         '''
         rows = ora_conn.run(sql, (username, table)).get_rows()
@@ -122,104 +104,28 @@ class OracleTools:
                     order by column_position 
                     rows between unbounded preceding and unbounded following
                 ) index_str
+                , uniqueness, index_type
                 from t_idx 
             '''
             rss = mdb.db_conn.run(sql).get_rows()
             mdb.close()
             for rs in rss:
-                mess = 'index: (' + rs[1] + ')'
-                tab = tab + mess + '\n'
-            
-        # 默认值
+                col = rs[1].lower()
+                if rs[3] == 'BITMAP':
+                    idx_type = ' bitmap'
+                elif rs[2] == 'UNIQUE':
+                    idx_type = ' unique'
+                else:
+                    idx_type = ''
+                ss = 'index: (' + col.replace(' asc', '').replace(',', ', ') + ')' + idx_type
+                mess = mess + ss + '\n'
+                
+        # 约束 类型 列和列排序
         sql = '''
-            select column_name, data_default
-            from dba_tab_columns
-            where owner = ? 
-            and table_name = ? 
-            and column_name not like '%$%'
-            and data_default is not null
-            order by column_id
-        '''
-        rss = ora_conn.run(sql, (username, table)).get_rows()
-        for rs in rss:
-            mess = 'data_default: ' + rs[0] + ' ' + rs[1].replace('  /* GOLDENGATE_DDL_REPLICATION */', '').strip().strip('\n')
-            tab = tab + mess + '\n'
-            
-        return tab
-    
-    @staticmethod
-    def get_ddl(dba_conn, username, table):
-        """ 得到目标user.table的单元性的建表语句，包括按顺序的列名、列类型、默认值、非空约束，普通索引、函数索引和注释 """
-        """ 不包括外键、分区信息、存储信息和特殊索引类型 """
-        username = username.upper()
-        table = table.upper()
-        rssqls = []
-        # 建表
-        rssqls.append('create table ' + username + '.' + table + '(test_col number)')
-        # 按顺序建列
-        sql = '''
-            select column_name, 
-                case 
-                    when data_type = 'VARCHAR2' or data_type = 'CHAR' or data_type = 'VARCHAR' then 
-                        data_type || '(' || data_length || ')'
-                    when data_type = 'NVARCHAR' or data_type = 'NVARCHAR2' then 
-                        data_type || '(' || char_length || ')'
-                    when data_type = 'NUMBER' and data_precision > 0 and data_scale > 0 then 
-                        data_type || '(' || data_precision || ', ' || data_scale || ')'
-                    when data_type = 'NUMBER' and data_precision > 0 and data_scale = 0 then 
-                        data_type || '(' || data_precision || ')'
-                    when data_type = 'NUMBER' and data_precision = 0 and data_scale = 0 then 
-                        data_type
-                    else data_type 
-                end column_type
-            from dba_tab_cols where owner = ? and table_name = ? and column_name not like '%$%' order by column_id
-        '''
-        for r in dba_conn.run(sql, (username, table)).get_rows():
-            rssqls.append('alter table ' + username + '.' + table + ' add ' + r[0] + ' ' + r[1].strip())
-        rssqls.append('alter table ' + username + '.' + table + ' drop column test_col')
-        # 建主键
-        sql = '''
-            select constraint_name
+            select constraint_name, constraint_type
             , max(cols)
             from (
-                select t.constraint_name
-                , to_char(
-                    wm_concat(t2.column_name) 
-                    over(partition by t.constraint_name order by t2.position)
-                ) cols
-                from dba_constraints t, dba_cons_columns t2
-                where t.owner = ?
-                and t.table_name = ?
-                and t.owner = t2.owner
-                and t.constraint_name = t2.constraint_name
-                and t.table_name = t2.table_name
-                and t.constraint_type = 'P'
-            ) group by constraint_name
-        '''
-        for r in dba_conn.run(sql, (username, table)).get_rows():
-            rssqls.append('alter table ' + username + '.' + table + ' add constraint ' + r[0] + ' primary key(' + r[1] + ')')
-        # 建非空约束
-        sql = '''
-            select t.search_condition
-            from dba_constraints t, dba_cons_columns t2
-            where t.owner = ?
-            and t.table_name = ?
-            and t.owner = t2.owner
-            and t.constraint_name = t2.constraint_name
-            and t.table_name = t2.table_name
-            and t.constraint_type = 'C'
-            and t.search_condition is not null
-        '''
-        for r in dba_conn.run(sql, (username, table)).get_rows():
-            if 'IS NOT NULL' in r[0]:
-                col = r[0].split(' ')[0]
-                rssqls.append('alter table ' + username + '.' + table + ' modify ' + col + ' not null')
-        # 建唯一约束
-        sql = '''
-            select constraint_name
-            , max(cols)
-            from (
-                select t.constraint_name
+                select t.constraint_name, t.constraint_type
                 , to_char(
                     wm_concat(t2.column_name)
                     over(partition by t.constraint_name order by t2.position)
@@ -230,95 +136,22 @@ class OracleTools:
                 and t.owner = t2.owner
                 and t.constraint_name = t2.constraint_name
                 and t.table_name = t2.table_name
-                and t.constraint_type = 'U'
-            ) group by constraint_name
+                and t.constraint_type in('P','U')
+            ) group by constraint_name, constraint_type
         '''
-        for r in dba_conn.run(sql, (username, table)).get_rows():
-            rssqls.append('alter table ' + username + '.' + table + ' add constraint ' + r[0] + ' unique(' + r[1] + ')')
-        # 建默认值
-        sql = '''
-            select column_name, data_default
-            from dba_tab_columns
-            where owner = ? 
-            and table_name = ? 
-            and column_name not like '%$%'
-            and data_default is not null
-            order by column_id
-        '''
-        for r in dba_conn.run(sql, (username, table)).get_rows():
-            rssqls.append('alter table ' + username + '.' + table + ' modify ' + r[0] + ' default ' + r[1].strip().strip('\n'))
-        # 建索引，包括普通索引和普通函数索引
-        sql = '''
-            select t.index_name
-            , t.index_type
-            , t2.column_name
-            , t3.column_expression
-            , t2.descend
-            , t2.column_position
-            from dba_indexes t, dba_ind_columns t2, dba_ind_expressions t3
-            where t.table_owner = ? 
-            and t.table_name = ? 
-            and t.owner = t2.index_owner
-            and t.table_owner = t2.table_owner
-            and t.index_name = t2.index_name
-            and t.table_name = t2.table_name
-            and t2.index_owner = t3.index_owner(+)
-            and t2.table_owner = t3.table_owner(+)
-            and t2.index_name = t3.index_name(+)
-            and t2.table_name = t3.table_name(+)
-            and t2.column_position = t3.column_position(+)
-            and t.uniqueness = 'NONUNIQUE'
-            and t.index_type in('FUNCTION-BASED NORMAL', 'NORMAL')
-            order by t2.column_position
-        '''
-        rows = dba_conn.run(sql, (username, table)).get_rows()
-        if len(rows) > 0:
-            col = dba_conn.run(sql, (username, table)).get_cols()
-            mdb = MemoryDB()
-            mdb.set('t_idx', col, rows)
-            # 降序索引也体现为函数索引，列名会用"col_name"，而字符串用'str'
-            sql = '''
-                select distinct index_name
-                , group_concat(
-                    replace(case when column_name like 'SYS_%$' then column_expression else column_name end, '"', '')
-                    || ' ' || descend
-                ) over(
-                    partition by index_name 
-                    order by column_position 
-                    rows between unbounded preceding and unbounded following
-                ) index_str
-                from t_idx 
-            '''
-            rs = mdb.db_conn.run(sql).get_rows()
-            mdb.close()
-            for r in rs:
-                rssqls.append('create index ' + username + '.' + r[0] + ' on ' + username + '.' + table + '(' + r[1] + ')')
-        # 建注释
-        sql = '''
-            select column_name, comments
-            from dba_col_comments
-            where owner = ? 
-            and table_name = ? 
-            and comments is not null
-        '''
-        for r in dba_conn.run(sql, (username, table)).get_rows():
-            rssqls.append("comment on column " + username + "." + table + "." + r[0] + " is '" + r[1] + "'")
-        # 建外键
-        pass
-        return rssqls
-    
-    @staticmethod
-    def get_ddl_by_dbms(dba_conn, username, table):
-        """ 得到目标user.table的的建表语句，直接调用dbms_metadata.get_ddl得到结果 """
-        username = username.upper()
-        table = table.upper()
-        sql = '''
-            select to_char(
-                dbms_metadata.get_ddl('TABLE', ?, ?)
-            ) from dual
-        '''
-        return dba_conn.run(sql, (table, username)).get_rows()[0]
-    
+        rss = ora_conn.run(sql, (username, table)).get_rows()
+        for rs in rss:
+            c_type = rs[1]
+            cols = rs[2].lower()
+            if c_type == 'U':
+                constraint_type = ' unique'
+            elif c_type == 'P':
+                constraint_type = ' pk'
+            else:
+                constraint_type = ''
+            mess = mess + 'constraint: (' + cols + ')' + constraint_type + '\n'
+        return mess.strip('\n')
+
     @staticmethod
     def compare_table(src_conn, src_username, src_table_name, dst_conn, dst_username, dst_table_name):
         """ 比较两个表，根据desc方法得到的字符串去比较 """
@@ -331,6 +164,41 @@ class OracleTools:
                                        , str_2=dst_username + '.' + dst_table_name + '\n' + dst_desc
                                        , iscompare=True) + '\n\n'
         return mess, src_desc == dst_desc
+  
+    @staticmethod
+    def get_ddl(dba_conn, username, table):
+        ddl_sqls = ''
+        sql = '''
+            select * from (
+                select table_owner owner, table_name, index_name obj_name, 'INDEX' obj_type from dba_indexes
+                union all
+                select owner, table_name, table_name, 'TABLE' obj_type from dba_tables
+            ) t where t.owner = ? and t.table_name = ?
+            order by decode(obj_type, 'TABLE', 0, 'INDEX', 1), obj_name
+        '''
+        rows = dba_conn.run(sql, (username, table)).get_rows()
+        for row in rows:
+            type_ = row[3]
+            name = row[2]
+            ddl_sql = OracleTools.get_ddl_by_dbms(dba_conn, username, type_, name) + ';'
+            ddl_sqls = ddl_sqls + '-- ' + type_ + ' ' + name + '\n'
+            ddl_sql = ddl_sql.lstrip().replace('"', '')
+            ddl_sqls = ddl_sqls + ddl_sql + '\n\n'
+            
+        # 建注释
+        sql = '''
+            select column_name, comments
+            from dba_col_comments
+            where owner = ? 
+            and table_name = ? 
+            and comments is not null
+        '''
+        
+        ddl_sqls = ddl_sqls + '-- 注释' + '\n'
+        for r in dba_conn.run(sql, (username, table)).get_rows():
+            ddl_sqls = ddl_sqls + "comment on column " + username + "." + table + "." + r[0] + " is '" + r[1] + "'\n"
+            
+        return ddl_sqls    
     
     @staticmethod
     def get_sids_by_host(host_conn):
@@ -342,7 +210,7 @@ class OracleTools:
         return sids
     
     @staticmethod
-    def get_obj_size(dba_conn):
+    def _get_obj_size(dba_conn):
         """ 得到对象(table、index、lob)大小 """
         sql = '''
             select (select utl_inaddr.get_host_address from dual) ip
@@ -371,7 +239,7 @@ class OracleTools:
         return dba_conn.run(sql)
     
     @staticmethod
-    def get_data_file_size(dba_conn):
+    def _get_data_file_size(dba_conn):
         """ 得到数据文件大小 """
         sql = '''
             select (select utl_inaddr.get_host_address from dual) ip
